
package fonctions ;
import java.lang.Math ;

/**
 * Cette classe définit la fonction représentant l'équation différentielle à 2N+2
 * dimensions du problème de la chaine de N carbones dont l'extrémité est fixée
 * à un bâti, modélisant un PAH et un hydrogène.
 * Il implémente l'interface Fonction, et appartient au package fonctions.
 */

public class FonctionChaineCFixee implements Fonction {

    /**
     * Nombre de carbones.
     */
    
    int N ;
	
    /**
     * Paramètre dé-dimensionné de masse.
     */
    
    double mu ;
    
    /**
     * Paramètre dé-dimensionné de l'énergie d'interaction entre le carbone et
     * l'hydrogène.
     */
    
    double V0 ;
    
    /**
     * Paramètre dé-dimensionné de la distance inter-atomes caractéristique dans 
     * une chaine carbonée.
     */
    
    double DCC ;

    /**
     * Constructeur de la classe.
     * @param N_syst Le paramètre N de la classe Systeme défini par l'utilisateur.
     * @param mu_syst Le paramètre mu de la classe Systeme défini par l'utilisateur.
     * @param V0_syst Le paramètre v0 de la classe Systeme défini par l'utilisateur.
     * @param DCC_syst Le paramètre DCC de la classe Systeme défini par l'utilisateur (dé-dimensionné).
     */
    
    public FonctionChaineCFixee(int N_syst, double mu_syst, double V0_syst, double DCC_syst) {  
    	this.N = N_syst ;   
    	this.mu = mu_syst ;
        this.V0 = V0_syst ;
        this.DCC = DCC_syst ;
    }
    
    /**
     * Implémentation de la méthode pour obtenir la dimension des tableaux 
     * auxquels s'applique la fonction.
     * @return Dimension des tableaux manipulés, ici 2*(N+1).
     */

    public int getDim() {
        return 2*N + 2 ;
    }

    /**
     * Implémentation de la fonction.
     * @param x Tableau auquel s'applique la fonction sous la forme 
     * (q_1, ..., q_n+1, p_1, ..., p_n+1).
     * @return Tableau résultat, la dérivée par rapport au temps de 
     * q_1, ..., p_1, ... à l'instant correspondant au tableau d'entrée.
     * Attention, alpha, paramètre de l'interaction, est supposé fixé et égal à 12.
     */

    public double[] f(double[] x) {

        double[] res = new double[2*N + 2] ;
        double[] q = new double[N + 1] ;
        double[] p = new double[N + 1] ;
        double[] q_point = new double[N + 1] ;
        double[] p_point = new double[N + 1] ;

        // Initialisation des variables.
        
        for(int i = 0 ; i <= N ; i++)
        {
            q[i] = x[i] ;
            p[i] = x[i + N + 1] ;
        }
        
        // Calcul du résultat.
        // Dérivées de (q_1, ..., q_n+1).
        // Attention, les tabs partent de l'indice 0, contrairement à q et p !!!
        
        for(int i = 0 ; i <= N-2 ; i++) // de q_1 à q_n-1
        {
            q_point[i] = p[i] ; // q_i+1 = p_i+1
        }
        q_point[N-1] = p[N-1] - p[N] ; // q_n = p_n - p_n+1
        q_point[N] = -p[N-1] + ( (1 + mu)/mu ) * p[N] ;
        
    	// Dérivées de (p_1, ..., p_n+1).
    	
    	if(N <= 1) // Les cas N<=0 n'ont pas de sens ici.
    	{
    	    p_point[0] = -q[0] + DCC ;
    	}
    	
    	else
    	{
    		p_point[0] = -2. * q[0] + q[1]  ; // p_1 = -2q_1 + q_2
    		for(int i = 1 ; i <= N-2 ; i++) // de p_2 à p_n-1
    		{
    			p_point[i] = q[i-1] - 2.*q[i] + q[i+1] ;
    		}
    		p_point[N-1] = q[N-2] - q[N-1] + DCC ;
    	}
    	
    	p_point[N] = 6. * V0 * ( 1. / Math.pow(q[N],13) - 1. / Math.pow(q[N],7) ) ;
    	
    	// Copie dans le tableau résultat.
    	
        for(int i = 0 ; i <= N ; i++)
        {
            res[i] = q_point[i] ;
            res[i + N + 1] = p_point[i] ;
        }   	
    	
        return(res) ;
    }
}
