
package integrateur ;
import fonctions.* ;
import java.lang.Math ;

/**
 * Contient des méthodes d'incrémentation
 * numérique de type Runge-Kutta, et uniquement des méthodes.
 * Elle fait partie du package integrateur, et ne peut fonctionner qu'avec le package fonctions.
 */

public class Increment{
    
    /**
     * rk4 est une méthode correspondant à  une incrémentation de type Runge-Kutta du quatrième ordre.
     * @param X Tableau d'entrée correspondant au vecteur avant incrémentation.
     * @param h Le pas d'incrémentation.
     * @param f La fonction correspondant à  l'équation différentielle. La dérivée de X(t) est f[X(t)].
     * @return Tableau résultat correspondant au vecteur incrémenté.
     */   
    public double[] rk4(double[] X, double h, Fonction f){
    
        /**
         * Récupération de la dimension des vecteurs manipulés par la fonction.
         */
        
        int n = f.getDim() ;

        /**
         * Calculs.
         */

        double[] k1 = new double[n] ;
        double[] k2 = new double[n] ;
        double[] k3 = new double[n] ;
        double[] k4 = new double[n] ;
        double[] Xtemp = new double[n] ;

        k1 = f.f(X) ;
        
        for(int i=0 ; i<n ; i++) {
            Xtemp[i] = X[i] + 0.5 * h * k1[i] ;
        }
        
        k2 = f.f(Xtemp) ;
        
        for(int i=0 ; i<n ; i++) {
            Xtemp[i] = X[i] + 0.5 * h * k2[i] ;
        }
        
        k3 = f.f(Xtemp) ;
        
        for(int i=0 ; i<n ; i++) {
            Xtemp[i] = X[i] + h * k3[i] ;
        }
        
        k4 = f.f(Xtemp) ;
        
        for(int i=0 ; i<n ; i++) {
            Xtemp[i] = X[i] + (1./6.) * h * (k1[i] + 2. * k2[i] + 2. * k3[i] + k4[i]) ;
        }
        
        return(Xtemp);
    }
    
    /**
     * rk4pv est une incrémentation de type Runge-Kutta du quatrième ordre avec pas variable.
     * @param X Xgen : vecteur systeme avant incrémentation.
     * @param Er double : L'erreur relative souhaitée sur l'incrementation.
     * @param dt double : Si l'erreur demandée requiert un pas plus grand que dt, alors on prend dt/3
     * @param f Fonction : La fonction correspondant à  l'équation différentielle. La dérivée de X(t) est f[X(t)].
     * @return Xgen : Tableau résultat correspondant au vecteur incrémenté.
     */   
    public Xgen rk4pv(Xgen X, double Er, double dt, Fonction f){
        
        // Initialisation des parametres nécessaires aux calculs.
        int imax, n = f.getDim() ;
        double c, h, h0, Eabs, Emax, dts3 ;
        double[] X1 = new double[n] ;
        double[] X2 = new double[n] ;
        double[] E = new double[n] ;
        Xgen res ;
        
        // Calcul du pas optimal h0.
        dts3 = dt / 3. ;
        res = X ;
        c = 1./15.0 ;
        h = X.geth() ;
        X1 = rk4(X.getX(), h, f) ;
        X2 = rk4( rk4(X.getX(), h/2.0, f ), h/2.0, f) ;
        
        for(int i=0 ; i<n ; i++) {
            E[i] = Math.abs( X1[i] - X2[i] ) ;
        }
        
        Emax = MinMax.maximum(E) ;
        imax = MinMax.getmaxindex(E) ;
        Eabs = Er * ( Math.abs(X2[imax]) + Math.abs( X.geth() * f.f(X2)[imax] ) ) ;
        h0 = 0.95 * h * Math.pow( Eabs / Emax, 1.0/5.0 ) ;
        

        // Si h0 est plus grand que dt/3, alors le nouveau pas utilisé est dt/3
        if( h0 > dts3 ) {
            h0 = dts3 ;
        }
        
        // Si l'erreur commise avec le pas h est superieure a l'erreur demandée,
        // on change le pas et on quitte la fonction.
        if( Emax > Eabs ) {
            res.seth(h0); 
        }

        // Sinon, on utilise une combinaison linéaire des deux rk4 calculés plus haut.
        else {
            for(int i=0 ; i<n ; i++) {
                res.setXi(( 16. * X2[i] - X1[i] ) * c, i) ;
            }
            res.seth(h0) ;
            res.sett(X.gett()+h) ;
        }
        return res ;
    }
    
    /**
     * rk4dt est un rk4 pas variable ou l'on peut fixer l'intervalle d'intégration.
     * @param X Xgen : vecteur systeme avant incrémentation.
     * @param dt double : La longueur de l'intervalle d'intégration.
     * @param Er double : L'erreur relative souhaitée sur l'incrementation.
     * @param f Fonction : La fonction correspondant à  l'équation différentielle. La dérivée de X(t) est f[X(t)].
     * @return Xgen : Tableau résultat correspondant au vecteur incrémenté.
     */   
    public double[] rk4dt(double[] X, double dt, double Er, Fonction f) {
    	
        // Initialisation des variables nécessaires.
        int n ;
        double t0, h0, c ;
        Xgen Xc = new Xgen(f.getDim(),1,0) ;
        Xgen Xtemp = new Xgen(f.getDim(),1,0) ;
        double[] res, x1, x2 ;

        Xc.setX(X) ;
        c = 1./15.0 ;
        n = f.getDim() ;
        res = new double[n] ;  x1 = new double[n] ;  x2 = new double[n] ;

        // On fait tourner rk4pv tant que le temps du système est inférieur à  tmax.
        while(Xc.gett() < dt) {
            Xtemp = Xc ;
            Xc = rk4pv(Xc, Er, dt, f) ;
        }
        
        h0 = dt-Xtemp.gett() ;
        x1 = rk4(Xtemp.getX(), h0, f) ;
        x2 = rk4(rk4(Xtemp.getX(), h0/2, f), h0/2, f) ;
        for(int i=0 ; i<n ; i++) {
                res[i] = (( 16. * x2[i] - x1[i] ) * c) ;
            }
        
        return res ;
    }
}
