package intergag;

/*
InterGag applet
Interaction grain-gaz
*/

import exceptions.*;
import simu.*;
import integrateur.* ;
import fonctions.* ;
import systeme.* ;
import java.util.* ;

/**
 * Algorithme de simulation pour SimuApplet.
 * Simule l'interaction d'un atome d'hydrogène avec une chaine d'atomes de carbone.
 * Permet l'affichage de différentes données intéressantes sur le problème.
 * compilation: javac -source 1.2 -target 1.1 -encoding UTF-8 -classpath SimuLab.jar Intergag.java
 */
public class Intergag extends SimuApplet {

    /**
     * Instance de la classe système qui stocke des données et méthodes utiles à notre problème
     */
    private Systeme systeme_intergag ;
    
    /**
     * Position initiale (en m)
     */
    private double x0 ;
    
    /**
     * Vitesse initiale (en m/s)
     */
    private double v0 ;

    /**
     * Distance caractéristique de l'interaction C-H (en m)
     */
    private double r0 ;
    
    /**
     * Energie du puits de potentiel (en J)
     */
    private double eps ;
    
    /**
     * Constante de raideur de l'interaction entre carbones (en kg/s²)
     */
    private double k ;
    
    /**
     * Masse d'un carbone (en kg)
     */
    private double M ;  
    
    /**
     * Nombre d'atomes de carbone
     */
    private int nc ;
    
    /**
     * Nombre d'itérations (appels à l'intégrateur)
     */
    private int n_iter ;
    
    /** 
     * Erreur maximale de l'intégrateur à pas variable sur une itération
     */
    private double errmax ;

    /**
     * Durée simulée dé-dimensionnée
     */
    private double tphys_max ;
    
    /**
     * Temps total minimal de l'affichage à respecter pour qu'il soit perceptible (en ms)
     * Le calcul est ralenti si l'affichage devait être trop rapide.
     */      
    private double temps_total_min ;
    
    /**
     * Choix de l'algorithme (pas variable ou fixe)
     */
    private String choixalgo ;
    
    /**
     * Choix du modèle pour la chaine de carbone
     */
    private String typechaine ;
    
    /**
     * Tableau de positions généralisées
     */
    private double qt[] ;
    
    /**
     * Tableau de moments conjugués
     */
    private double pt[] ;
    
    /**
     * Masse réduite (dé-dimensionné)
     */
    private double mu ;  
    
    /**
     * Potentiel caractéristique de l'interaction C-H dé-dimensionné V0
     */
    private double potV0 ;
    
    /**
     * Distance caractéristique inter-carbones dé-dimensionnée
     */
    private double dcc ;
    
    /** 
     * Temps physique de la simulation dé-dimensionné
     */
    private double tphys ;

    /**
     * Ecart de temps physique dé-dimensionné entre deux itérations.
     */
    private double delta_tphys ;
    
    /**
     * Temps de la machine au début du calcul (en ms)
     */
    private double tmachine_init ;

    /**
     * Temps de la machine à l'instant courant (en ms)
     */    
    private double tmachine_courant ;

    /**
     * Compteur d'itérations
     */      
    private int compteur_iter ;
    
    /**
     * Pas fixe, si utilisé (temps dé-dimensionné)
     */
    private double pasfixe ;
    
    /** 
     * Erreur sur l'énergie totale du système, qui devrait être conservée : (E-E0) / E0
     */
    private double err_energie ;
    
    
    /**
     * Constructeur de la classe, avec la définition des paramètres de simulation
     * et des options d'affichage.
     */
    public Intergag() {
        appletDoc("intergag_doc.xml");
    }


    /**
     * Calcul live (utilisé avec affichage plotlive)
     * Initialisation
     */    
    public void initCalculLive(ListeValeurs in) throws SimuException {

        // Récupération des données saisies par l'utilisateur
        //x0 = in.lireDouble("x0") * 1E-9 ; // Converti en m depuis les nm.
        v0 = in.lireDouble("v0") ; // Gardé en m/s.
        eps = in.lireDouble("eps") * 1.6E-19 ; // Converti en J depuis les eV.
        r0 = in.lireDouble("r0") * 1.E-9 ; // Converti en m depuis les nm.
        nc = in.lireEntier("nc") ;
        tphys_max = in.lireDouble("tmax") * 1.E-12 ; // Passé en s depuis les ps, à dé-dimensioner.
        errmax = in.lireDouble("errmax") ;
        n_iter = in.lireEntier("n_iter") ;
        temps_total_min = in.lireDouble("temps_total_min") * 1.E3 ; // Converti en ms depuis les s.
        choixalgo = in.lireString("choixalgo") ;
        typechaine = in.lireString("typechaine") ;
        
        Systeme systeme_dcc ;
        try {
        	systeme_dcc = new Systeme(3,50.E-9,-300.,0.07* 1.6E-19,1.E-9,Systeme.cc_fixe) ; // Pour pouvoir récupérer dcc.
        }
        catch(ParamException err_param)
        {
            SimuException err = new SimuException(err_param.getMessage() + " Erreur sur \"" + err_param.getNomParam() + "\"") ;
            throw err ;
        }
        
        if(typechaine.equals("fixee")) {  // Maintenant on fixe x0 à 5nm du carbone le plus proche, cela ne vient plus de l'utilisateur.
        	x0 = 5.0E-9 + nc * systeme_dcc.getDcc() ;
        }
        else {
        	x0 = 5.0E-9 + nc * systeme_dcc.getDcc() / 2. ;
        }
        
        if ( (nc == 1) && typechaine.equals("libre") ) {
        	SimuException my_err = new SimuException("Il faut au moins 2 carbones pour le cas de la chaîne libre !") ;
            throw my_err ;
        }
        
        try
        {
            // Initialisation du système
            if(typechaine.equals("fixee")) {
                systeme_intergag = new Systeme(nc,x0,v0,eps,r0,Systeme.cc_fixe) ;
            }
            else {
                systeme_intergag = new Systeme(nc,x0,v0,eps,r0,Systeme.cc_libre) ;
            }
        }
        
        catch(ParamException err_param)
        {
            SimuException err = new SimuException(err_param.getMessage() + " Erreur sur \"" + err_param.getNomParam() + "\"") ;
            throw err ;
        }

        // Récupération des propriétés du système
        
        k = systeme_intergag.getK() ;
        mu = systeme_intergag.getMu() ;
        M = systeme_intergag.getMasseC() ;
        potV0 = systeme_intergag.getV0() ;
        dcc = systeme_intergag.getDcc() / r0 ; // dé-dimensionné !
        
        // Dé-dimensionnement de tphys_max et calculs des temps
        
        tphys_max *= Math.sqrt(k/M) ;
        
        delta_tphys = tphys_max / n_iter ;
        tphys = 0. ;
        pasfixe = delta_tphys ;
        compteur_iter = 0 ;

        // Initialisation de qt et pt

        qt = new double[nc + 1] ;
        pt = new double[nc + 1] ;
        
        for(int i=0; i<=nc; i++) {
            qt[i] = systeme_intergag.getVar()[i] ;
            pt[i] = systeme_intergag.getVar()[i + nc + 1] ;
        }
       
        // EXTENSION DE L'AFFICHAGE
        
        // Ajustement des paramètres de sortie
        Vector vect_param_sortie = getOutParams() ;
      
        // Destruction des éléments du vecteur de paramètre de sortie au-delà de 6
        // Si nc = 1, aucune action
        for(int j = 0; j<getOutParams().size() - 6; j++) {
        
            vect_param_sortie.removeElementAt(j + 6) ;
        
        }
        
        // Ajustement de la taille
        vect_param_sortie.setSize(4 + nc*2) ;
      
        // Complément avec le nombre de paramètres souhaité
        // Si nc = 1, aucune action
        
        for(int j=0; j<nc - 1; j++) {
                
                ParamOut temp_param= new ParamOut("nombre", "xc" + (j+2), "", "") ;
                ParamOut temp_param2= new ParamOut("nombre", "yc" + (j+2), "", "") ;
                vect_param_sortie.setElementAt(temp_param, 2*j + 6) ;
                vect_param_sortie.setElementAt(temp_param2, 2*j + 7) ;

        }
        
        // Création d'un objet affichage et récupération de l'affichage
        Affichage trajanim = getAffichage("trajanim") ;
        
        // Récupération du 1er sous-affichage
        Affichage traj = trajanim.sousAff[0] ;
        
        // Récupération du nombre de paramètres du 1er sous-affichage
        int longueur_tableau = 2 ;
        
        String[][] newparams = new String[longueur_tableau - 1 + nc][2] ; 
         
        // Recopie du tableau de paramètres
        for(int j=0; j<longueur_tableau; j++) {
            for(int i=0;i<2;i++) {
            newparams[j][i] = traj.params[j][i] ;
            }
        }
        
        
        // Ajout des autres paramètres
        for(int j=longueur_tableau; j<longueur_tableau - 1 + nc; j++) {
        
            for(int i=0; i<2; i++) {
            
                if (i==0) {
                    newparams[j][i] = "t" ;
                }
                else {
                    newparams[j][i] = "xc" + j ;
                }
            
            }
        }
        
        // Mise à jour de la liste des paramètres du sous-affichage
        traj.params = newparams ;
        
                        
        // Récupération du 2ème sous-affichage
        Affichage traj2 = trajanim.sousAff[1];
        
        
        // Récupération du nombre de paramètres du 2ème sous-affichage
        int longueur_tableau2 = 2 ;
        
        String[][] newparams2 = new String[longueur_tableau - 1 + nc][2] ; 
        
        // Recopie du tableau de paramètres
        for(int j=0; j<longueur_tableau2; j++) {
            for(int i=0; i<2; i++) {
            	newparams2[j][i] = traj2.params[j][i] ;
            }
        }
        
        // Ajout des autres paramètres
        for(int j=longueur_tableau2; j<longueur_tableau2 - 1 + nc; j++) {
        
            for(int i=0; i<2; i++) {
            
                if (i == 0) {
                    newparams2[j][i] = "yc" + j ;
                }
                else {
                    newparams2[j][i] = "xc" + j ;
                }
            
            }
        }
        
        // Mise à jour de la liste des paramètres du sous-affichage
        traj2.params = newparams2 ;



        // Création d'un objet affichage et récupération de l'affichage
        Affichage trajerr = getAffichage("trajerr") ;
        
        // Récupération du 1er sous-affichage
        Affichage trajerr1 = trajerr.sousAff[0] ;
                
        // Récupération du nombre de paramètres du 2ème sous-affichage
        int longueur_tableau3 = 2 ;

        String[][] newparams3 = new String[longueur_tableau3 - 1 + nc][2] ; 
         
        // Recopie du tableau de paramètres
        for(int j=0; j<longueur_tableau3; j++) {
            for(int i=0;i<2;i++) {
            newparams3[j][i] = trajerr1.params[j][i] ;
            }
        }
        
        
        // Ajout des autres paramètres
        for(int j=longueur_tableau3; j<longueur_tableau3 - 1 + nc; j++) {
        
            for(int i=0; i<2; i++) {
            
                if (i==0) {
                    newparams3[j][i] = "t" ;
                }
                else {
                    newparams3[j][i] = "xc" + j ;
                }
            
            }
        }
        
        // Mise à jour de la liste des paramètres du sous-affichage
        trajerr1.params = newparams3 ;

        
        // Récupération du temps de l'ordinateur au début de la simulation
        tmachine_init = System.currentTimeMillis() ;
        
    }

    
    /**
     * Calcul live (utilisé avec affichage plotlive)
     * Boucle
     */   
    public ListeValeurs calculLive() throws SimuException {

        if (choixalgo.equals("pas variable")) {

        	// Condition de fin de calcul
        	
            if (tphys >= tphys_max)
                return null ;
            
            // Création d'une liste de valeurs pour affichage
            
            ListeValeurs out = new ListeValeurs() ;

            // var_avant contient dans l'ordre q1, q2, ..., q_nc+1, p1, p2    , ..., p_nc+1 
            
            double var_avant[] = new double[2*(nc+1)] ;   
            double var_apres[] = new double[2*(nc+1)] ;
            double pos_hydrogene ;
            double vit_hydrogene ;
            double pos_carbone[] = new double[nc] ;

            // Initialisation de var_avant
            
            for(int i=0; i<=nc; i++) {
            	var_avant[i] = qt[i] ;
            	var_avant[i + nc + 1] = pt[i] ;
            }
 
            // Création de la fonction f de type FonctionChaineCFixee ou FonctionChaineCLibre
            // qui modélise le hamiltonien du problème à intégrer
        
            Fonction f_fix = new FonctionChaineCFixee(nc, mu, potV0, dcc) ;
            Fonction f_lib = new FonctionChaineCLibre(nc, mu, potV0, dcc) ;
        
            // Création d'un intégrateur
            
            Increment incr = new Increment() ;
        
            // Appel à l'intégrateur Runge-Kutta d'ordre 4 à pas variable 
            
            if(typechaine.equals("fixee")) {
            	var_apres = incr.rk4dt(var_avant, delta_tphys, errmax, f_fix) ;
            }
            else {
            	var_apres = incr.rk4dt(var_avant, delta_tphys, errmax, f_lib) ;
            }
   
            // Mise à jour du vecteur
            
            for(int i=0; i<=nc; i++) {
            	qt[i] = var_apres[i] ;
            	pt[i] = var_apres[i + nc + 1] ;
            }

            // Calcul de l'erreur sur l'énergie totale du système
            
            int modele_chaine = 0 ;
            if(typechaine.equals("fixee")) {
            	modele_chaine = 1 ;
            }
            err_energie = systeme_intergag.getErrEnergie(var_apres, modele_chaine) ;
        
            // Re-dimensionnement des positions et vitesses
            
            pos_hydrogene = 1.E9 * r0 * ( qt[nc-1] + qt[nc] ) ; // position de l'hydrogène en nm
            vit_hydrogene = r0 * pt[nc] * Math.sqrt(k/M) / mu ; // vitesse de l'hydrogène en m/s
        
            for(int i=0; i<=nc-1; i++) {     
            	pos_carbone[i] = 1.E9 * r0 * qt[i] ; // position du carbone i+1 en nm (le carbone 1 est le plus loin de l'hydrogène)
            }
            // vitesse des carbones non codée.

            // Mise à jour de la liste de valeurs pour affichage
            
            out.ajouterDouble("t", Math.sqrt(M/k) * tphys) ; // Temps re-dimensionné
            out.ajouterDouble("xt", pos_hydrogene) ;
            out.ajouterDouble("yt", 0) ; // Obligatoire pour debuger l'affichage

            // Ajout des positions des n carbones pour affichage
            
            for(int i=0; i<nc; i++) {
            	out.ajouterDouble("xc" + (i+1), pos_carbone[nc - 1 - i]) ; // Renvoie la position du carbone i+1 pour l'affichage
            	out.ajouterDouble("yc" + (i+1), 0) ; // Obligatoire pour debuger l'affichage
            }

            out.ajouterDouble("err_energie", err_energie) ; // Erreur sur l'énergie totale   
    
            // Ralentissement si nécessaire du calcul pour que l'affichage soit perceptible.
            
            tmachine_courant = System.currentTimeMillis() ;
            
            if( tmachine_courant < tmachine_init + temps_total_min * compteur_iter / n_iter) {
            	dormir((long) (tmachine_init + temps_total_min * compteur_iter / n_iter  - tmachine_courant) ) ;
            }
    
            // Incrémentation du temps physique et du compteur d'incrémentations.
            
            tphys = tphys + delta_tphys ;
            compteur_iter ++ ;
  
            return out ;
        }
        
        // Intégration à pas fixe
        else   
        {
        	// Condition de fin de calcul
        	
        	if (compteur_iter >= n_iter)
        		return null ;

            // Création d'une liste de valeurs pour affichage
            
        	ListeValeurs out = new ListeValeurs() ;

            // var_avant contient dans l'ordre q1, q2, ..., q_nc+1, p1, p2    , ..., p_nc+1 
            
            double var_avant[] = new double[2*(nc+1)] ;   
            double var_apres[] = new double[2*(nc+1)] ;
            double pos_hydrogene ;
            double vit_hydrogene ;
            double pos_carbone[] = new double[nc] ;

            // Initialisation de var_avant
            
            for(int i=0; i<=nc; i++) {
            	var_avant[i] = qt[i] ;
            	var_avant[i + nc + 1] = pt[i] ;
            }
 
            // Création de la fonction f de type FonctionChaineCFixee ou FonctionChaineCLibre
            // qui modélise le hamiltonien du problème à intégrer
        
            Fonction f_fix = new FonctionChaineCFixee(nc, mu, potV0, dcc) ;
            Fonction f_lib = new FonctionChaineCLibre(nc, mu, potV0, dcc) ;
        
            // Création d'un intégrateur
            
            Increment incr = new Increment() ;
        
            // Appel à l'intégrateur Runge-Kutta d'ordre 4 à pas fixe
            
        	if(typechaine.equals("fixee")) {
        		var_apres = incr.rk4(var_avant, pasfixe, f_fix) ;
        	}
        	else {
        		var_apres = incr.rk4(var_avant, pasfixe, f_lib) ;
        	}
   
            // Mise à jour du vecteur
            
            for(int i=0; i<=nc; i++) {
            	qt[i] = var_apres[i] ;
            	pt[i] = var_apres[i + nc + 1] ;
            }

            // Calcul de l'erreur sur l'énergie totale du système
            
            int modele_chaine = 0 ;
            if(typechaine.equals("fixee")) {
            	modele_chaine = 1 ;
            }
            err_energie = systeme_intergag.getErrEnergie(var_apres, modele_chaine) ;
        
            // Re-dimensionnement des positions et vitesses
            
            pos_hydrogene = 1.E9 * r0 * ( qt[nc-1] + qt[nc] ) ; // position de l'hydrogène en nm
            vit_hydrogene = r0 * pt[nc] * Math.sqrt(k/M) / mu ; // vitesse de l'hydrogène en m/s
        
            for(int i=0; i<=nc-1; i++) {     
            	pos_carbone[i] = 1.E9 * r0 * qt[i] ; // position du carbone i+1 en nm (le carbone 1 est le plus loin de l'hydrogène)
            }
            // vitesse des carbones non codée.

            // Mise à jour de la liste de valeurs pour affichage
            
            out.ajouterDouble("t", Math.sqrt(M/k) * tphys) ; // Temps re-dimensionné
            out.ajouterDouble("xt", pos_hydrogene) ;
            out.ajouterDouble("yt", 0) ; // Obligatoire pour debuger l'affichage

            // Ajout des positions des n carbones pour affichage
            
            for(int i=0; i<nc; i++) {
            	out.ajouterDouble("xc" + (i+1), pos_carbone[nc - 1 - i]) ; // Renvoie la position du carbone i+1 pour l'affichage
            	out.ajouterDouble("yc" + (i+1), 0) ; // Obligatoire pour debuguer l'affichage
            }

            out.ajouterDouble("err_energie", err_energie) ; // Erreur sur l'énergie totale   
    
            // Ralentissement si nécessaire du calcul pour que l'affichage soit perceptible.
            
            tmachine_courant = System.currentTimeMillis() ;
            
            if( tmachine_courant < tmachine_init + temps_total_min * compteur_iter / n_iter) {
            	dormir((long) (tmachine_init + temps_total_min * compteur_iter / n_iter  - tmachine_courant) ) ;
            }
    
            // Incrémentation du temps physique et du compteur d'incrémentations.
            
            tphys = tphys + delta_tphys ;
            compteur_iter ++ ;
     
        	    return out;      
        }
    }
        
}

