package systeme;
import exceptions.ParamException;
import java.lang.Math;
import java.lang.System;


/**
 * Définit le système choisi par l'utilisateur pour le projet Intergag et charge
 * tous les paramètres associés.
 * [ Par défaut collision C-H ]
 */
public class Systeme
{   
    /**
     * Tableau de particules, instances de la classe Particule.
     * Attention, l'hydrogène est le premier élément du tableau (position 0).
     */
    private Particule[] tab_particules ;
    
    /**
     * Nombre d'atomes de carbone.
     */
    private int nc ;

    /**
     * Position initiale de l'atome d'hydrogène.
     */
    private double xh ;

    /**
     * Vitesse initiale de l'atome d'hydrogène.
     */
    private double vh ;

    /**
     * Distance caractéristique du potentiel d'interaction.
     */
    private double r0 ;

    /**
     * Profondeur du potentiel d'interaction (en Joules).
     */
    private double eps ;

    /**
     * Constante de raideur du "ressort" liant les carbones.
     */
    private double k ;
    
    /**
     * Potentiel d'interaction dé-dimensionné.
     */
    private double v_0 ;
    
    /**
     * Masse dé-dimensionnée.
     */
    private double mu ;

    /**
     * Tableau des variables réduites générées.
     */
    private double[] var_gen ;
    
    /**
     * Energie totale initiale du système.
     */
    private double h0 ;

    /**
     * Energie totale au temps t du système.
     */
    private double h ; 
    
    /**
     * Erreur sur l'énergie totale : (E(t)-E0) / E0.
     */   
    private double err_energie ;

    /**
     * Valeur de la fréquence de vibration typique des carbones (en Hz).
     */
    private static final double fvib = 1.e12 ;

    /**
     * Longueur à vide du "ressort" : distance C-C (en mètres).
     */
    private static final double dcc = 150.e-12 ;
    
    /**
     * Booléen pour le cas de la chaine de carbone libre.
     */
    public static final int cc_libre = 0 ;
    
    /**
     * Booléen pour le cas de la chaine de carbone fixée à un grain.
     */   
    public static final int cc_fixe = 1 ;
    
    // /**
    //  * Cas du benzène.
    //  */    
    // public static final int benz = 2 ;
    
    // /**
    //  * Nombre de dimensions du problème
    //  */
    //  private int dim = 1 ;
    
    
    /**
     * Constructeur par initialisation.
     * Types de particules : h = hydrogène, c = carbone.
     * @param nb_c Nombre de carbones
     * @param pos position hydrogène, à donner en m.
     * @param vit vitesse hydrogène, à donner en m/s.
     * @param eps profondeur du puits, à donner en Joules
     * @param r0 distance caractéristique du potentiel, à donner en mètre
     * @param model précise le modèle choisi pour la chaine de carbone, libre(0) ou fixée (1)
     */
    public Systeme(int nb_c, double pos, double vit, double eps, double r0, int model) throws ParamException
    {

    	// Recopie et stocke les valeurs données par l'utilisateur.
        this.eps = eps ;
        this.r0 = r0 ;
        nc = nb_c ;
        xh = pos ;
        vh = vit ;
        
        tab_particules = new Particule[nc+1];
        
        // Création de l'hydrogène.
        tab_particules[0] = new Particule(Particule.h, xh, vh);
        
        // Création des carbones selon le modèle de la chaine de carbone choisi.
        if(model == cc_libre)
        {
            
            // Vérification que l'hydrogène est bien positionné plus loin que la chaine.
            if(xh <= dcc * (nc-1) /2. )
            {
                ParamException erreur = new ParamException("L'hydrogène doit être placé plus loin que la chaîne de carbone !\n\n","Position de l'hydrogène");
                throw erreur ;
            }
            
            /* Condition sur la vitesse de l'hydrogène (doit être négative) */
            /*if(vh > 0)
            {
                ParamException erreur = new ParamException("La vitesse de l'hydrogène doit être négative !\n\n","Vitesse de l'hydrogène");
                throw erreur;
            }*/
            
            for(int i = 1; i <= nc; i++)
            {
                // Barycentre du système de carbones origine des abscisses.
                tab_particules[i] = new Particule(Particule.c, dcc * (2*i-(nc+1)) / 2., 0) ;
            }
            
        }
        else if(model == cc_fixe)
        {
            // Vérification que l'hydrogène est bien positionné plus loin que la chaine.
            if(xh <= dcc * nc )
            {
                ParamException erreur = new ParamException("L'hydrogène doit être placé plus loin que la chaîne de carbone !\n\n","Position de l'hydrogène");
                throw erreur ;
            }
            
            /* Condition sur la vitesse de l'hydrogène (doit être négative) */
            /*if(vh > 0)
            {
                ParamException erreur = new ParamException("La vitesse de l'hydrogène doit être négative !\n\n","Vitesse de l'hydrogène");
                throw erreur;                
            }*/
            
            for(int i = 1; i <= nc; i++)
            {
            // Bâti origine des abscisses.
            tab_particules[i] = new Particule(Particule.c, i * dcc, 0) ;
            }
            
        }
        
        // Calcul de mu, V0 et k.
        mu = tab_particules[0].getMasse() / tab_particules[1].getMasse() ;
        k = 4. * Math.pow(Math.PI, 2) * Math.pow(fvib, 2) * tab_particules[1].getMasse() ;
        v_0 = 2. * eps / ( k * Math.pow(r0, 2) ) ;
           
        // Tableau des variables réduites (format (q1,q2,...,p1,p2,...)
        var_gen = new double[2 * (nc+1)] ;
        for(int i = 0; i < nc-1; i++) // du carbone 1 au carbone N-1.
        {
            // Positions.
            var_gen[i] = tab_particules[i+1].getPosition() / r0 ;  // Position du carbone i+1.
            // Impulsions.
            var_gen[i+1+nc] = tab_particules[i+1].getVitesse() * Math.sqrt( tab_particules[i+1].getMasse() / k ) / r0 ; // Vitesse du carbone i+1.
        }
        
        // Carbone N.
        var_gen[nc-1] = tab_particules[nc].getPosition() / r0 ;
        var_gen[2*nc] = ( ( tab_particules[nc].getVitesse() + mu * tab_particules[0].getVitesse() ) * Math.sqrt( tab_particules[nc].getMasse() / k ) ) / r0 ;
        
        // Hydrogène.
        var_gen[nc] = ( tab_particules[0].getPosition() - tab_particules[nc].getPosition() ) / r0 ;  // Position H - position C(nc) (carbone en bout de chaine).
        var_gen[2*nc+1] = ( mu * tab_particules[0].getVitesse() * Math.sqrt( tab_particules[nc].getMasse() / k ) ) / r0 ; //Impulsion nc+1.
        
        // Calcul de l'énergie initiale du système.
        
    	h0 = 0. ; // Initialisation.
    	
    	for (int i = 0; i <= nc-2; i++) { // Somme de l'énergie dans les carbones 1 à nc-1, quel que soit le cas.
    	    h0 += Math.pow(var_gen[nc+1+i], 2) / 2. + Math.pow(var_gen[i+1] - var_gen[i] - dcc/r0, 2) / 2. ;
    	}
    		
    	h0 += Math.pow(var_gen[2*nc] - var_gen[2*nc+1], 2) / 2. + Math.pow(var_gen[2*nc+1], 2) / (2.*mu) ; // Energie cinétique dans le dernier carbone et l'hydrogène.
    	
    	h0 += v_0 * (0.5 / Math.pow(var_gen[nc],12) - 1. / Math.pow(var_gen[nc],6) ) ; // Energie potentielle d'interaction.
    	
    	if(model == cc_fixe) {
    		h0 += Math.pow(var_gen[0] - dcc/r0, 2) / 2. ; // Energie contenue dans le ressort fixé au bâti pour le cas fixe.
    	}
    	
    }
    
    
    
    /**
     * Méthode accédant à la variable V0.
     * @return V_0
     */
    public double getV0()
    {
        return v_0 ;
    }
    
    /**
     * Méthode accédant à la variable mu.
     * @return mu
     */   
    public double getMu()
    {
        return mu ;
    }
    
    /**
     * Méthode accédant à la variable k.
     * @return k
     */   
    public double getK()
    {
        return k ;
    }
    
    /**
     * Méthode accédant à la masse du carbone.
     * @return masse du carbone
     */
    public double getMasseC()
    {
        return tab_particules[1].getMasse() ;
    }
    
    /**
     * Méthode accédant au tableau des variables réduites.
     * @return tableau des variables réduites.
     */   
    public double[] getVar()
    {
        return var_gen ;
    }
    
    /**
     * Méthode accédant à la distance carbone-carbone (longueur à vide du ressort).
     * @return distance carbone-carbone en m.
     */
    public double getDcc()
    {
        return dcc ;
    }   
    
    
    /**
     * Méthode calculant l'erreur sur l'énergie au temps considéré.
     * @param var Tableau des variables réduites à l'instant t
     * @param model Modèle choisi pour la chaine de carbone, 0 pour libre et 1 pour fixée.
     * @return erreur sur l'énergie au temps t
     */
    public double getErrEnergie(double[] var, int model)
    {
        
    	h = 0. ; // Initialisation.
    	
    	for (int i = 0; i <= nc-2; i++) { // Somme de l'énergie dans les carbones 1 à nc-1, quel que soit le cas.
    	    h += Math.pow(var[nc+1+i], 2) / 2. + Math.pow(var[i+1] - var[i] - dcc/r0, 2) / 2. ;
    	}
    		
    	h += Math.pow(var[2*nc] - var[2*nc+1], 2) / 2. + Math.pow(var[2*nc+1], 2) / (2.*mu) ; // Energie cinétique dans le dernier carbone et l'hydrogène.
    	
    	h += v_0 * (0.5 / Math.pow(var[nc],12) - 1. / Math.pow(var[nc],6) ) ; // Energie potentielle d'interaction.
    	
    	if(model == cc_fixe) {
    		h += Math.pow(var[0] - dcc/r0, 2) / 2. ; // Energie contenue dans le ressort fixé au bâti pour le cas fixe.
    	}
        
        err_energie = Math.abs( (h-h0) / h0 );
        
        return err_energie ;
    }
}
